"""Adaptors for getting useful information out of bzrlib objects."""

from StringIO import StringIO
from textwrap import dedent
from webob import html_escape

from bzrlib.repository import Repository
from bzrlib.branch import Branch
from bzrlib.log import LogFormatter
from bzrlib.osutils import format_date

class BazaarRepositoryAdaptor(object):
    """Wrap a bzrlib repository object and expose useful information
    about it.
    """

    def __init__(self, repository):
        if not isinstance(repository, Repository):
            raise TypeError(dedent("""\
                BazaarRepositoryAdaptor requires an instance of
                bzrlib.repository.Repository.
                """))
        self.repo = repository
        self.branches = [BazaarBranchAdaptor(b) 
                    for b in self.repo.find_branches()]

    def get_branches(self):
        """Get a list of branch adaptor objects for all branches in this
        repository.
        """
        return self.branches
            
    def get_branch(self, branch_name):
        """Get a branch adaptor object for a single, named branch.
        """
        for b in self.branches:
            if b.get_name() == branch_name:
                return b
        


class BazaarBranchAdaptor(object):
    """Wrap a bzrlib Branch object and expost useful information about
    it.
    """

    def __init__(self, branch):
        if not isinstance(branch, Branch):
            raise TypeError(dedent("""\
                BazaarBranchAdaptor requires an instance of
                bzrlib.branch.Branch.
                """))
        self.branch = branch

    def get_name(self):
        """Get the branch name.

        In fact this returns the branch nickname, since the branch name
        seems to always be set to None
        """
        return html_escape(self.branch.get_config().get_nickname())

    def get_log(self):
        """Get a log of this branch.
        """
        from bzrlib.log import Logger, make_log_request_dict

        buf = StringIO()
        log_formatter = HTMLLogFormatter(to_file=buf)
        # limit to 10 revisions.
        request = make_log_request_dict(limit=10, levels=1)
        Logger(self.branch, request).show(log_formatter)

        return buf.getvalue()
        
    def get_last_revision_number(self):
        """
        Get the last revision number for this branch, or 0 if there are no
        revisions.
        """
        return self.branch.last_revision_info()[0]


class HTMLLogFormatter(LogFormatter):
    """A log formatter that outputs HTML.
    """

    supports_delta = True
    supports_merge_revisions = True
    supports_tags = True
    supports_diff = True

    def __init__(self, *args, **kwargs):
        super(HTMLLogFormatter, self).__init__(*args, **kwargs)

        self._line_no =1

    def log_revision(self, revision):
        revno = revision.revno
        authors = revision.rev.get_apparent_authors()
        message = revision.rev.message
        dt = format_date(revision.rev.timestamp,
            revision.rev.timezone or 0,
            self.show_timezone)
        row_class = ''
        if self._line_no % 2 == 0:
            row_class='even'

        self.to_file.write(dedent("""\
        <tr class="revision %s">
            <td class="revno">%s</td>
            <td class="date">%s</td>
            <td class="authors">%s</td>
            <td class="message">%s</td>
        </tr>
        """ % (row_class,
            revno,
             dt,
            html_escape(", ".join(authors)),
            html_escape(message))))
        self._line_no += 1

    def begin_log(self):
        self.to_file.write(dedent('''\
        <table class="bzr-log">
        <tr class="heading">
            <td>Rev#</td>
            <td>Date</td>
            <td>Authors</td>
            <td>Commit Message</td>
        </tr>
        '''))

    def end_log(self):
        self.to_file.write('</table>')

    def get_levels(self):
        return 1










