"""Functions relating to the handling of bazaar repositories and branches."""

from datetime import datetime
import os
from shutil import copytree, rmtree

from bzrlib import bzrdir, transport
from bzrlib.repository import Repository
from sloecode.bzr.display import BazaarRepositoryAdaptor


def ensure_directory_exists(dir_name):
    """Make a directory if it doesn't already exist.
    
    Throws RuntimeError if the path already exists as something other
    than a directory (a file, for example).
    """
    if os.path.exists(dir_name):
        if not os.path.isdir(dir_name):
            raise RuntimeError("%s exists but isn't a directory", dir_name)
    else:
        os.makedirs(dir_name)


def create_shared_repo(dir_name):
    """Create a Bazaar repository at the given location.
    """
    format = bzrdir.format_registry.make_bzrdir('default')
    to_transport = transport.get_transport(dir_name)
    to_transport.ensure_base()
    newdir = format.initialize_on_transport(to_transport)
    repo = newdir.create_repository(shared=True)
    repo.set_make_working_trees(False)
    return repo


class BazaarRepositoryFactory(object):
    """Responsible for creating repositories for people and projects."""

    def __init__(self, person_root_dir, project_root_dir, 
    deleted_person_dir, deleted_project_dir):
        self.person_root_dir = person_root_dir
        self.project_root_dir = project_root_dir
        self.deleted_person_dir = deleted_person_dir
        self.deleted_project_dir = deleted_project_dir
        
        ensure_directory_exists(self.person_root_dir)
        ensure_directory_exists(self.project_root_dir)
        ensure_directory_exists(self.deleted_person_dir)
        ensure_directory_exists(self.deleted_project_dir)

    def get_user_repo_dir(self, person_name):
        """
        Get the directory for a users repository, given the username.
        """
        return os.path.join(self.person_root_dir, person_name)
    
    def get_project_repo_dir(self, project_name):
        """
        Get the directory for a project repository, given the project name.
        """
        return os.path.join(self.project_root_dir, project_name)
    
    def create_shared_repository_for_person(self, person_name):
        """Create an empty shared repo for person."""
        dir_name = self.get_user_repo_dir(person_name)
        create_shared_repo(dir_name)

    def create_shared_repository_for_project(self, project_name):
        """Create a repository for the project with a trunk branch."""
        dir_name = self.get_project_repo_dir(project_name)
        repo = create_shared_repo(dir_name)
        # Now to create the trunk.
        to_transport = repo.user_transport.clone('trunk')
        format = bzrdir.format_registry.make_bzrdir('default')
        branch = bzrdir.BzrDir.create_branch_convenience(
            to_transport.base, format=format, 
            possible_transports=[to_transport])
        branch.set_append_revisions_only(True)
        
    def get_repository_for_project(self, project_name):
        """Get a bzrlib.repository.Repository object for a given project
        name.
        """
        dir_name = self.get_project_repo_dir(project_name)
        return BazaarRepositoryAdaptor(Repository.open(dir_name))
        
    def get_repository_for_person(self, person_name):
        """
        Get a bzrlib.repository.Repository object for a given person login.
        """
        dir_name = self.get_user_repo_dir(person_name)
        return BazaarRepositoryAdaptor(Repository.open(dir_name))
    
    def delete_shared_repository_for_person(self, person_name):
        """
        Delete a personal repository. This function should only be called 
        AFTER the user account has been deleted. This function will create a
        copy of the users repository in the personal bzr repo backup folder.
        """
        src_dir = self.get_user_repo_dir(person_name)
        dst_dir = self.get_delete_folder_for_user(person_name)
        
        copytree(src_dir, dst_dir)
        rmtree(src_dir)
        
    def delete_shared_repository_for_project(self, project_name):
        """
        Delete a project repository. This function should only be called 
        AFTER the project has been deleted. This function will create a
        copy of the project repository in the project repo backup folder.
        """
        src_dir = self.get_project_repo_dir(project_name)
        dst_dir = self.get_delete_folder_for_project(project_name)
        
        copytree(src_dir, dst_dir)
        rmtree(src_dir)
        
    def get_delete_folder_for_user(self, person_name):
        """
        Get a path to a directory that does not yet exist, suitable for backing
        up a user repository.
        
        Directories are in the form <username>-<timestamp>, allowing
        administrators to easily differentiate between multiple user backups over
        time.
        """
        dt = datetime.utcnow()
        dirname = person_name + '-' + dt.strftime("%Y-%M-%d_%H%M%S")
        dirpath = os.path.join(self.deleted_person_dir, dirname)
        attempt = 0
        while (os.path.exists(dirpath)):
            if attempt != 0:
                dirpath = dirpath[:-2]
            dirpath += '_' + str(attempt)
            attempt += 1
            
        return dirpath
    
    def get_delete_folder_for_project(self, project_name):
        """
        Get a path to a directory that does not yet exist, suitable for backing
        up a user repository.
        
        Directories are in the form <username>-<timestamp>, allowing
        administrators to easily differentiate between multiple user backups over
        time.
        """
        dt = datetime.utcnow()
        dirname = project_name + '-' + dt.strftime("%Y-%M-%d_%H%M%S")
        dirpath = os.path.join(self.deleted_project_dir, dirname)
        attempt = 0
        while (os.path.exists(dirpath)):
            if attempt != 0:
                dirpath = dirpath[:-2]
            dirpath += '_' + str(attempt)
            attempt += 1
            
        return dirpath
