# Copyright 2009 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Bazaar plugin to run the smart server for Sloecode.

Cribbed from bzrlib.builtins.cmd_serve from Bazaar 0.16.
"""

__metaclass__ = type

__all__ = [
    'cmd_sloecode_server',
    ]


import resource
import sys

from bzrlib import lockdir, ui
from bzrlib.commands import Command, register_command
from bzrlib.option import Option
from bzrlib.smart import medium, server
from bzrlib.transport import get_transport


class cmd_sloecode_server(Command):
    """Run a Bazaar server that maps branch URLs to the underlying sloecode
    directories for user and projects.
    """

    aliases = ['sc-serve']

    takes_options = [
        Option('inet',
               help='serve on stdin/out for use from inetd or sshd'),
        Option('port',
               help='listen for connections on nominated port of the form '
                    '[hostname:]portnumber. Passing 0 as the port number will'
                    ' result in a dynamically allocated port. Default port is'
                    ' 4022.',
               type=str),
        Option('user-directory',
               help='The directory that holds the user repositories.',
               type=unicode),
        Option('project-directory',
               help='The directory that holds the project repositories.'),
        Option('server-endpoint',
               help='The url of the XML-RPC server.',
               type=unicode),
        ]

    takes_args = ['username']

    def get_smart_server(self, transport, port, inet):
        """Construct a smart server."""
        if inet:
            smart_server = medium.SmartServerPipeStreamMedium(
                sys.stdin, sys.stdout, transport)
        else:
            host = medium.BZR_DEFAULT_INTERFACE
            if port is None:
                port = medium.BZR_DEFAULT_PORT
            else:
                if ':' in port:
                    host, port = port.split(':')
                port = int(port)
            smart_server = server.SmartTCPServer(transport)
            smart_server.start_server(host, port)
            print 'listening on port: ', smart_server.port
            sys.stdout.flush()
        return smart_server

    def run_server(self, smart_server):
        """Run the given smart server."""
        # for the duration of this server, no UI output is permitted.
        # note that this may cause problems with blackbox tests. This should
        # be changed with care though, as we dont want to use bandwidth
        # sending progress over stderr to smart server clients!
        old_factory = ui.ui_factory
        try:
            ui.ui_factory = ui.SilentUIFactory()
            smart_server.serve()
        finally:
            ui.ui_factory = old_factory

    def run(self,
            username,
            port=None,
            inet=False,
            user_directory=None,
            project_directory=None,
            server_endpoint=None):
        from sloecode.codehosting.server import get_sc_server
        from sloecode.lib.readconf import get_conf
        # TODO: check the relativeness of the path being passed in.
        # The current working directory will almost certainly be wrong.
        config = get_conf('smartserver')
        if server_endpoint is None:
            server_endpoint = config['xmlrpc_auth_endpoint']
        if user_directory is None:
            user_directory = config['bzr_person_root']
        if project_directory is None:
            project_directory = config['bzr_project_root']
        if port is None:
            port = int(config['ssh_port_number'])
        four_gig = int(4e9)
        resource.setrlimit(resource.RLIMIT_AS, (four_gig, four_gig))
        lp_server = get_sc_server(
            username,
            user_directory,
            project_directory,
            server_endpoint)
        lp_server.start_server()

        old_lockdir_timeout = lockdir._DEFAULT_TIMEOUT_SECONDS
        try:
            lp_transport = get_transport(lp_server.scheme)
            smart_server = self.get_smart_server(lp_transport, port, inet)
            lockdir._DEFAULT_TIMEOUT_SECONDS = 0
            self.run_server(smart_server)
        finally:
            lockdir._DEFAULT_TIMEOUT_SECONDS = old_lockdir_timeout
            lp_server.stop_server()


register_command(cmd_sloecode_server)


def load_tests(standard_tests, module, loader):
    # None at the moment.
    return standard_tests
