# Copyright 2009 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

# pylint: disable-msg=W0702,W0231

"""Transport utilities for the codehosting system.

The code hosting filesystem is implemented using Bazaar transports. This
module contains utilities for implementing virtual filesystems using
bzrlib.transport classes.
"""

__metaclass__ = type
__all__ = [
    'VirtualTransport',
    'get_chrooted_transport',
    'get_readonly_transport',
    'TranslationError',
    ]


from bzrlib import urlutils
from bzrlib.errors import (
    BzrError,
    InProcessTransport,
    TransportNotPossible,
    )
from bzrlib.transport import (
    chroot,
    get_transport,
    Transport,
    )


class TranslationError(BzrError):
    """Raised when we cannot translate a virtual URL fragment.

    In particular, this is raised when there is some intrinsic deficiency in
    the path itself.
    """

    _fmt = ("Could not translate %(virtual_url_fragment)r. %(reason)s")

    def __init__(self, virtual_url_fragment, reason=None):
        BzrError.__init__(self)
        self.virtual_url_fragment = virtual_url_fragment
        if reason is not None:
            self.reason = str(reason)
        else:
            self.reason = ''


def get_chrooted_transport(url, mkdir=False):
    """Return a chrooted transport serving `url`."""
    transport = get_transport(url)
    if mkdir:
        transport.create_prefix()
    chroot_server = chroot.ChrootServer(transport)
    chroot_server.start_server()
    return get_transport(chroot_server.get_url())


def get_readonly_transport(transport):
    """Wrap `transport` in a readonly transport."""
    if transport.base.startswith('readonly+'):
        return transport
    return get_transport('readonly+' + transport.base)


class VirtualTransport(Transport):
    """A Sloecode virtual transport that knows about the users and projects."""

    def __init__(self, server, url):
        self.server = server
        Transport.__init__(self, url)

    def external_url(self):
        # There's no real external URL to this transport. It's heavily
        # dependent on the process.
        raise InProcessTransport(self)

    def _abspath(self, relpath):
        """Return the absolute, escaped path to `relpath` without the schema.
        """
        return urlutils.joinpath(
            self.base[len(self.server.scheme)-1:], relpath)

    def _getUnderylingTransportAndPath(self, relpath):
        """Return the underlying transport and path for `relpath`."""
        virtual_url_fragment = self._abspath(relpath)
        return self.server.translateVirtualPath(virtual_url_fragment)

    def _call(self, method_name, relpath, *args, **kwargs):
        """Call a method on the backing transport, translating relative,
        virtual paths to filesystem paths.

        If 'relpath' translates to a path that we only have read-access to,
        then the method will be called on the backing transport decorated with
        'readonly+'.
        """
        transport, path = self._getUnderylingTransportAndPath(relpath)
        method = getattr(transport, method_name)
        return method(path, *args, **kwargs)

    # Transport methods
    def abspath(self, relpath):
        return urlutils.join(self.base, relpath)

    def append_file(self, relpath, f, mode=None):
        return self._call('append_file', relpath, f, mode)

    def clone(self, relpath=None):
        if relpath is None:
            return self.__class__(self.server, self.base)
        else:
            return self.__class__(
                self.server, urlutils.join(self.base, relpath))

    def delete(self, relpath):
        return self._call('delete', relpath)

    def delete_tree(self, relpath):
        return self._call('delete_tree', relpath)

    def get(self, relpath):
        return self._call('get', relpath)

    def get_bytes(self, relpath):
        return self._call('get_bytes', relpath)

    def has(self, relpath):
        return self._call('has', relpath)

    def iter_files_recursive(self):
        transport, path = self._getUnderylingTransportAndPath('.')
        return transport.clone(path).iter_files_recursive()

    def listable(self):
        transport, path = self._getUnderylingTransportAndPath('.')
        return transport.listable()

    def list_dir(self, relpath):
        return self._call('list_dir', relpath)

    def lock_read(self, relpath):
        return self._call('lock_read', relpath)

    def lock_write(self, relpath):
        return self._call('lock_write', relpath)

    def mkdir(self, relpath, mode=None):
        return self._call('mkdir', relpath, mode)

    def open_write_stream(self, relpath, mode=None):
        return self._call('open_write_stream', relpath, mode)

    def put_file(self, relpath, f, mode=None):
        return self._call('put_file', relpath, f, mode)

    def local_realPath(self, relpath):
        # This method should return an absolute path (not URL) that points to
        # `relpath` and dereferences any symlinks. The absolute path should be
        # on this transport.
        #
        # Here, we assume that the underlying transport has no symlinks
        # (Bazaar transports cannot create symlinks). This means that we can
        # just return the absolute path.
        return self._abspath(relpath)

    def readv(self, relpath, offsets, adjust_for_latency=False,
              upper_limit=None):
        return self._call(
            'readv', relpath, offsets, adjust_for_latency, upper_limit)

    def rename(self, rel_from, rel_to):
        to_transport, to_path  = self._getUnderylingTransportAndPath(rel_to)
        from_transport, from_path = self._getUnderylingTransportAndPath(rel_from)
        if to_transport.base != from_transport.base:
            raise TransportNotPossible(
                'cannot move between underlying transports')
        return getattr(from_transport, 'rename')(from_path, to_path)

    def rmdir(self, relpath):
        return self._call('rmdir', relpath)

    def stat(self, relpath):
        return self._call('stat', relpath)

    def writeChunk(self, relpath, offset, data):
        return self._call('writeChunk', relpath, offset, data)
