"""Controller for non-admin views on the person object.
"""
import logging
from pylons import request, config
from pylons.controllers.util import redirect, abort
from repoze.what.plugins.pylonshq import ActionProtector
from repoze.what.predicates import Any
from routes import url_for
import formencode

from repoze.what.predicates import not_anonymous

from sloecode.lib.base import BaseController
from sloecode.lib.predicates import has_site_role, owns_page
from sloecode.model.site_role import USER_ADMIN
from sloecode.model.person import Person
from sloecode.model.authkey import AuthKey, AuthKeySchema
from sloecode.model.meta import Session
import sloecode.lib.helpers as h


log = logging.getLogger(__name__)


class PersonController(BaseController):
    """PersonController renders pages for the person model.
    """

    @ActionProtector(Any(has_site_role(role=USER_ADMIN), owns_page()))
    def index(self, person_name):
        """Render the person details page.
        """
        ppl = Person.get(login=person_name)
        person = ppl[0] if ppl else None
        if person:
            repo = config['bzr.factory'].get_repository_for_person(person_name)
            return self.render('/person-details.html',
                {'person' : person,
                'repo': repo})
        else:
            abort(404)

    @ActionProtector(Any(has_site_role(role=USER_ADMIN), owns_page()))
    def branch_log(self, person_name):
        """Render a branch log for a personal repo branch.
        """
        ppl = Person.get(login=person_name)
        person = ppl[0] if ppl else None
        if person:
            repo = config['bzr.factory'].get_repository_for_person(person_name)
            branch_name = request.params['branch_name']
            branch = repo.get_branch(branch_name)
            return self.render('/person-branch-log.html',
                {'person' : person,
                'branch': branch})
        else:
            abort(404)

    # allow anonymous action, since all we do is redirect to a page with
    # checks:
    @ActionProtector(Any(has_site_role(role=USER_ADMIN), not_anonymous()))
    def me(self):
        """Redirect to the main user page.
        """
        user = request.environ['repoze.who.identity'].get('user')
        return redirect(url_for(controller='/person', action='index',
                                person_name=user.login))

    @ActionProtector(Any(has_site_role(role=USER_ADMIN), owns_page()))
    def manage_ssh_keys(self, person_name):
        """Render page that displays stored SSH keys, and allows user to add
        and remove them.
        """
        ppl = Person.get(login=person_name)
        person = ppl[0] if ppl else None
        return self.render('/person-manage-keys.html', {'person' : person})

    @ActionProtector(Any(has_site_role(role=USER_ADMIN), owns_page()))
    def process_key_add(self, person_name):
        ppl = Person.get(login=person_name)
        person = ppl[0]

        schema = AuthKeySchema()
        try:
            form_result = schema.to_python(request.params)
        except formencode.Invalid, e:
            errors = e.unpack_errors()
            return formencode.htmlfill.render(
                self.manage_ssh_keys(person_name), defaults=request.params, errors=errors)

        kt, kd, kc = form_result['ssh_key']
        name = form_result['ssh_key_name']
        person.auth_keys.append(AuthKey(key_type=kt, key_data=kd,
                                        key_comment=kc, key_name=name))
        Session.commit()
        h.flash("SSH Key added Successfully.")
        return redirect(url_for(controller='/person', action='manage_ssh_keys',
                                person_name=person.login))

    @ActionProtector(Any(has_site_role(role=USER_ADMIN), owns_page()))
    def process_key_del(self, person_name):
        ppl = Person.get(login=person_name)
        person = ppl[0]
        id = int(request.params['id'])
        keys = AuthKey.get(id=id)
        key = keys[0] if keys else None

        person.auth_keys.remove(key)
        Session.delete(key)
        Session.commit()

        h.flash("SSh Key Removed Successfully.")
        return redirect(url_for(controller='/person', action='manage_ssh_keys',
                                person_name=person.login))
