"""DB Model for storing authentication keys. Probably SSH keys.
"""

from sqlalchemy import Column
from sqlalchemy.types import Integer, Text
from sqlalchemy.schema import ForeignKey
import formencode
import re

from sloecode.model.meta import Base, QueryMixin


class AuthKey(Base, QueryMixin):
    "A model that holds authentication keys for a user."
    __tablename__ = "person_keys"

    id = Column(Integer, primary_key=True)
    person_id = Column(ForeignKey('person.id'))

    # Human friendly name to identify key
    key_name = Column(Text)

    # key type. Will most likely be 'ssh-rsa' or similar.
    key_type = Column(Text)

    # key data.
    # TODO: how do we tell SQlAlchemy to make this NOT NULL?
    key_data = Column(Text)

    # optional comment for the key.
    key_comment = Column(Text)


class KeyValidator(formencode.FancyValidator):
    """A simple formencode validator that checks that an SSH key is valid, and
    splits it into it's component parts.
    """

    # Regular expression to split apart SSH keys into type, data, comment.
    # Comment component is optional:
    regexp = re.compile(r'(\S+)\s+(\S+)\s*(\S*)')

    def _to_python(self, value, state):
        """Try to validate form data.
        """
        match = self.regexp.match(value)
        if match:
            return match.groups()
        else:
            raise formencode.Invalid(
                'Invalid SSH Key.', value, state)

class AuthKeySchema(formencode.Schema):
    "A formencode schema for the AuthKey object."

    allow_extra_fields = True
    filter_extra_fields = True

    ssh_key = KeyValidator(not_empty=True, strip=True)
    ssh_key_name = formencode.validators.String(not_empty=True)
