#!/usr/bin/env python

from sqlalchemy.exc import IntegrityError
from unittest import TestCase

from sloecode.model.meta import Session, Base, engine
from sloecode.model import Project, Person, Membership

class TestProject(TestCase):
    def setUp(self):
        # create all DB tables:
        Base.metadata.create_all(engine)

    def tearDown(self):
        # remove all DB tables.
        Base.metadata.drop_all(engine)
        # remove session - avoids a memory leak, apparently.
        Session.remove()

    def test_create(self):
        "can we create a project?"

        p1 = Project(name='Test project 1')
        self.failIf(p1 is None)

    def test_unique(self):
        "Test to make sure the project name is unique."

        p1 = Project(name='myAwesomeProject')
        Session.add(p1)
        Session.commit()

        p2 = Project(name='myAwesomeProject')
        Session.add(p2)
        self.failUnlessRaises(IntegrityError, Session.commit)

    def test_cascading_project_delete(self):
        '''Verify bug #724089: Deleting projects does not delete the
        records in the membership table.
        '''
        sc = Project(name='sloecode', description='')
        me = Person(login='thomir', name='thomas')
        mem = Membership(person=me, project=sc)

        Session.add(me)
        Session.add(sc)
        Session.add(mem)
        Session.commit()

        self.assertTrue(me in Session)
        self.assertTrue(mem in Session)
        self.assertTrue(sc in Session)

        # deleting the project should remove all memberships
        Session.delete(sc)
        Session.commit()

        # I should still exist!
        self.assertTrue(me in Session)

        self.assertTrue(mem not in Session)
        self.assertTrue(sc not in Session)

