#!/usr/bin/env python

from unittest import TestCase
from sqlalchemy.exc import IntegrityError

from sloecode.model.meta import Session, Base, engine
from sloecode.model import Person, Project, Membership, SiteRole



class TestRoles(TestCase):
    """Test membership roles model.
    """

    def setUp(self):
        # create all DB tables:
        Base.metadata.create_all(engine)

    def tearDown(self):
        # remove all DB tables.
        Base.metadata.drop_all(engine)
        # remove session - avoids a memory leak, apparently.
        Session.remove()

    def test_simpleRole(self):
        """Ensure we can give a users all three different roles."""
        # create a simple project and a few users:
        p = Project(name='sloecode',
                description='Simple Project Hosting')
        Session.add(p)
        me = Person(login='thomi', name='Thomi Richards',
            email='thomir@gmail.com')
        Session.add(me)
        tim =  Person(login='tim', name='Tim Penhey',
            email='tim@somewhere.com')
        Session.add(tim)
        rachel = Person(login='rachel', name='Rachel Saunders',
            email='rachel@somewhere.com')
        Session.add(rachel)

        Session.commit()

        # make me an observer:
        Session.add(Membership(person=me, project=p, role='Observer'))
        # make tim a developer:        
        Session.add(Membership(person=tim, project=p, role='Developer'))
        # make rachel a manager:
        Session.add(Membership(person=rachel, project=p, role='Manager'))
        Session.commit()

        # make sure we can't add other types of roles that have not been
        # defined in the roles model:
        Session.add(Membership(person=me, project=p, role='god'))
        self.failUnlessRaises(IntegrityError, Session.commit)
        Session.rollback()  # rollback that error above:

        # I am a member of exactly one project:
        self.assertTrue(len(me.projects) == 1)
        #... which is sloecode:
        self.assertTrue(me.projects[0].project.name == 'sloecode')
        # which is an observer:
        self.assertTrue(me.projects[0].role == u'Observer')

        # Similar tests for Tim:
        # Tim is a member of exactly one project:
        self.assertTrue(len(tim.projects) == 1)
        #... which is sloecode:
        self.assertTrue(tim.projects[0].project.name == 'sloecode')
        # which is a developer:
        self.assertTrue(tim.projects[0].role == u'Developer')

        # Similar tests for Rachel:
        # Rachel is a member of exactly one project:
        self.assertTrue(len(rachel.projects) == 1)
        #... which is sloecode:
        self.assertTrue(rachel.projects[0].project.name == 'sloecode')
        # which is 'Manager'
        self.assertTrue(rachel.projects[0].role == u'Manager')

    def test_site_roles(self):
        "Make sure site-wide roles work as expected."
        me = Person(login='thomi', name='Thomi', email='thomir@gmail.com')
        Session.add(me)
        
        tim = Person(login='tim', name='Tim', email='tim@somewhere.com')
        Session.add(tim)
        
        Session.commit()
        
        self.assertTrue(len(me.site_roles) == 0)
        self.assertTrue(len(tim.site_roles) == 0)
        
        # make me a user admin:
        Session.add(SiteRole(person=me, role='User Admin'))
        # make tim a project admin:
        Session.add(SiteRole(person=tim, role='Project Admin'))
        Session.commit()
        
        self.assertTrue(len(me.site_roles) == 1)
        self.assertTrue(len(tim.site_roles) == 1)

        # make me a project admin as well:
        Session.add(SiteRole(person=me, role='Project Admin'))
        Session.commit()
        self.assertTrue(len(me.site_roles) == 2)
        
        # get all user admins:
        ppl_roles = SiteRole.get(role='User Admin')
        self.assertTrue(len(ppl_roles) == 1)
        self.assertTrue(ppl_roles[0].person is me)
